"""
Implements RegEx abstraction logic, for easy working with RegEx in Python.
"""
import sys

from .interface import (
    after,
    any_of,
    before,
    char_in,
    char_not_in,
    create,
    exactly,
    not_after,
    not_before,
    one_or_more,
    optional,
    ref,
    zero_or_more,
)

if sys.version_info >= (3, 8):  # pragma: no cover
    from importlib import metadata as importlib_metadata
else:
    import importlib_metadata  # pragma: no cover


def _metadata():
    return importlib_metadata.metadata(__name__)


__metadata__ = _metadata()


def get_version() -> str:
    try:
        return str(__metadata__["version"])
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

__title__ = __metadata__["name"]
__version__ = version
__author__ = __metadata__["author"]
__license__ = __metadata__["licence"]
__copyright__ = "Copyright 2022 Ificiana"

CHAR = create(r".")
WHITESPACE = WS = create(r"\s")
NOT_WHITESPACE = NOT_WS = create(r"\S")
WORD = create(r"\w")
NOT_WORD = create(r"\W")
DIGIT = create(r"\d")
NOT_DIGIT = create(r"\D")
LETTER = create(r"[a-zA-Z]")
NOT_LETTER = create(r"[^a-zA-Z]")
TAB = create(r"\t")
NEWLINE = N = create(r"\n")
NOT_NEWLINE = NOT_N = create(r"\N")
CARRIAGE_RETURN = R = create(r"\r")

__all__ = [
    "CARRIAGE_RETURN",
    "CHAR",
    "DIGIT",
    "LETTER",
    "N",
    "NEWLINE",
    "NOT_DIGIT",
    "NOT_LETTER",
    "NOT_N",
    "NOT_NEWLINE",
    "NOT_WHITESPACE",
    "NOT_WORD",
    "NOT_WS",
    "R",
    "TAB",
    "WHITESPACE",
    "WORD",
    "WS",
    "after",
    "any_of",
    "before",
    "char_in",
    "char_not_in",
    "create",
    "exactly",
    "not_after",
    "not_before",
    "one_or_more",
    "optional",
    "version",
    "get_version",
    "zero_or_more",
    "ref",
]



from typing import Tuple, Union

import warnings

from .exceptions import RemagicException

try:  # pragma: no cover
    import regex as re
except ImportError:  # pragma: no cover
    import re  # type: ignore[no-redef]

    warnings.warn("`regex` module not found, using builtin `re` module", ImportWarning)


class Pattern:
    _pattern: str

    def __init__(self, pattern=""):
        self._pattern = pattern

    def compile(self) -> "re.Pattern":
        return re.compile(str(self._pattern))

    def add(self, other: Union[str, "Pattern"]) -> "Pattern":
        """
        Function to add two Patterns
        :param other: Pattern
        :return: other Pattern appended to this Pattern
        """
        return self + other

    def repeat(self, num: Union[int, Tuple[int, int]]) -> "Pattern":
        return self * num

    def label(self, name):
        return Pattern(rf"(?P<{name}>{self._pattern})")

    def group(self):
        return Pattern(rf"({self._pattern})")

    def ref(self, reference):
        if isinstance(reference, int) and reference > 0:
            return self.add(rf"\{reference}")
        if isinstance(reference, str):
            return self.add(rf"(?P={reference})")
        raise RemagicException("Only positive integers and strings are allowed")

    @property
    def pattern(self) -> str:
        return self._pattern  # pragma: no cover

    @pattern.getter
    def pattern(self) -> str:
        return self._pattern

    @pattern.setter
    def pattern(self, value):
        self._pattern = value

    def __add__(self, other: Union["Pattern", str]) -> "Pattern":
        if isinstance(other, Pattern):
            return Pattern(self._pattern + other._pattern)
        return Pattern(self._pattern + str(other))

    def __mul__(self, num: Union[int, Tuple[int, ...]]) -> "Pattern":
        if isinstance(num, tuple):
            for i in num:
                if i < 0:
                    raise RemagicException("Value should be positive")
            if len(num) == 1:
                (num,) = num
                if num == 0:
                    temp = "*"
                elif num == 1:
                    temp = "+"
                else:
                    temp = f"{{{num},}}"
            else:
                if len(num) > 2:
                    raise RemagicException("More than 2 values provided")
                first, second = num[0], num[1]
                if num[0] > num[1]:
                    raise RemagicException("Second value cannot be less than the first")
                if first == 0 and second == 1:
                    temp = "?"
                elif first == second:
                    temp = f"{{{first}}}"
                else:
                    temp = f"{{{first},{second}}}"
            return Pattern(self._pattern + temp)
        if num < 0:
            raise RemagicException("Value should be positive")
        return Pattern(self._pattern + f"{{{int(num)}}}")

    def __eq__(self, other):
        return self._pattern == str(other)

    def __iter__(self):
        return iter(self._pattern)  # pragma: no cover

    def __str__(self) -> str:
        return str(self._pattern)



from .exceptions import RemagicException
from .pattern import Pattern
from .utils import char_set_escape, escape


def create(pattern):
    if isinstance(pattern, Pattern):
        return pattern
    return exactly(pattern)


def exactly(pattern):
    return Pattern(pattern)


def optional(pattern):
    return Pattern(pattern) * (0, 1)


def zero_or_more(pattern, greedy=True):
    if greedy:
        return Pattern(pattern) * (0,)
    return optional(Pattern(pattern) * (0,))


def one_or_more(pattern, greedy=True):
    if greedy:
        return Pattern(pattern) * (1,)
    return optional(Pattern(pattern) * (1,))


def _char_set_items(iterable):
    # ordered set of items
    return dict.fromkeys(list("".join([char_set_escape(i) for i in iterable])))


def _alt_items(iterable):
    # ordered set of items
    return dict.fromkeys([escape(i) for i in iterable])


def char_in(iterable):
    return exactly(f"[{''.join(_char_set_items(iterable))}]")


def char_not_in(iterable):
    return exactly(f"[^{''.join(_char_set_items(iterable))}]")


def any_of(iterable):
    return exactly(f"{'|'.join(_alt_items(iterable))}")


# TODO: Write tests for `before`, `not_before`, `after`, `not_after`


def before(pattern):  # pragma: no cover
    return rf"(?={create(pattern)})"


def not_before(pattern):  # pragma: no cover
    return rf"(?!={create(pattern)})"


def after(pattern):  # pragma: no cover
    return rf"(?<={create(pattern)})"


def not_after(pattern):  # pragma: no cover
    return rf"(?<!={create(pattern)})"


def ref(reference):
    if isinstance(reference, int) and reference > 0:
        return exactly(rf"\{reference}")
    if isinstance(reference, str):
        return exactly(rf"(?P={reference})")
    raise RemagicException("Only positive integers and strings are allowed")



import re

SPECIALS = re.compile(r"([.+*?^$()\[\]{}|\\])")
CLASS_SPECIALS = re.compile(r"([\^\-\]\\])")


def escape(string):
    return SPECIALS.sub(r"\\\1", str(string))


def char_set_escape(string):
    return CLASS_SPECIALS.sub(r"\\\1", str(string))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['remagic']

package_data = \
{'': ['*']}

extras_require = \
{':extra == "regex"': ['regex>=2022.9.13,<2023.0.0'],
 ':python_version < "3.8"': ['importlib_metadata>=4.5.0,<5.0.0']}

setup_kwargs = {
    'name': 'remagic',
    'version': '0.1.2a0',
    'description': 'Working with RegEx made easy!',
    'long_description': '[![Coverage Status](https://coveralls.io/repos/github/ificiana/remagic/badge.svg?branch=main)](https://coveralls.io/github/ificiana/remagic?branch=main)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/remagic)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/remagic)\n![PyPI - License](https://img.shields.io/pypi/l/remagic)\n\n# remagic\n\nWorking with regex made easier!\nPartly inspired from `magic-regexp` for Node\n\n```py\nfrom remagic import *\n\npattern1 = create(DIGIT)  # matches any digit\npattern2 = char_in("aeiou")  # matches any char in "aeiou"\npattern3 = ~pattern2  # negates the pattern in pattern2, \n# i.e. match everything except characters in "aeiou"\n# Note: remagic 0.1.1 doesn\'t support the ~ syntax\n# use char_not_in("aeiou") instead\npattern4 = pattern1 + pattern3\n# finally compile, use standard flags as optional argument\nR = pattern4.compile()\n# use the regex later\n``` \n\n## Installation\n\nInstall from PyPI:\n`pip install remagic`\n\n## Work in Progress!\n\n### Documentation\n\nTODO: [docs](https://ificiana.github.io/remagic)\n\n### Known bugs\n\n- improper behaviour with `any_of`\n',
    'author': 'Ificiana',
    'author_email': 'ificiana@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ificiana/remagic',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
